# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import os
import sys
import time
import pickle
import subprocess as sp

import threading as THR
import queue as QU

import numpy as np
import pandas as pd


class OdbServerError(Exception):
    pass


class OdbClient:

    def __init__(self, abaqus_bin, python_env_path, odb_file):
        self._proc = None
        env = os.environ
        env['PYTHONPATH'] = self._guess_pythonpath(python_env_path)

        lock_file_exists = os.path.isfile(os.path.splitext(odb_file)[0] + '.lck')

        self._proc = sp.Popen([abaqus_bin, 'python', '-m', 'odbserver', odb_file],
                              stdout=sp.PIPE, stdin=sp.PIPE, stderr=sp.PIPE,
                              env=env)

        if lock_file_exists:
            self._gulp_lock_file_warning()

        self._wait_for_server_ready_sign()

    def _guess_pythonpath(self, python_env_path):
        if sys.platform == 'win32':
            return os.path.join(python_env_path, 'lib', 'site-packages')
        return os.path.join(python_env_path, 'lib', 'python2.7', 'site-packages')

    def _gulp_lock_file_warning(self):
            self._proc.stdout.readline()
            self._proc.stdout.readline()

    def _wait_for_server_ready_sign(self):
        def wait_for_input(stdout, queue):
            sign = stdout.read(5)
            queue.put(sign)

        queue = QU.Queue()
        thread = THR.Thread(target=wait_for_input, args=(self._proc.stdout, queue))
        thread.daemon = True
        thread.start()

        while True:
            self._check_if_process_still_alive()
            try:
                sign = queue.get_nowait()
            except QU.Empty:
                time.sleep(1)
            else:
                if sign != b'ready':
                    raise OdbServerError("Expected ready sign from server, received %s" % sign)
                return

    def instances(self):
        return _decode_ascii_list(self._query('get_instances'))

    def nodes(self, instance_name, node_set_name=b''):
        index, node_data = self._query('get_nodes', (instance_name, node_set_name))
        return pd.DataFrame(data=node_data, columns=['x', 'y', 'z'],
                            index=pd.Int64Index(index, name='node_id'))

    def connectivity(self, instance_name, element_set_name=b''):
        index, connectivity = self._query('get_connectivity', (instance_name, element_set_name))
        return pd.DataFrame({'connectivity': connectivity},
                            index=pd.Int64Index(index, name='element_id'))

    def node_sets(self, instance_name=b''):
        return _decode_ascii_list(self._query('get_node_sets', instance_name))

    def node_set(self, node_set_name, instance_name=b''):
        node_set = self._query('get_node_set', (instance_name, node_set_name))
        return pd.Int64Index(node_set, name='node_id')

    def element_sets(self, instance_name=b''):
        return _decode_ascii_list(self._query('get_element_sets', instance_name))

    def element_set(self, element_set_name, instance_name=b''):
        element_set = self._query('get_element_set', (instance_name, element_set_name))
        return pd.Int64Index(element_set, name='element_id')

    def steps(self):
        return self._query('get_steps')

    def frames(self, step_name):
        return self._query('get_frames', step_name)

    def variable_names(self, step, frame):
        return _decode_ascii_list(self._query('get_variable_names', (step, frame)))

    def variable(self, instance, step, frame, var_name, node_set_name=b'', element_set_name=b''):
        response = self._query('get_variable', (instance, step, frame, var_name, node_set_name, element_set_name))
        (labels, index_labels, index_data, values) = response

        index_labels = _decode_ascii_list(index_labels)
        if len(index_labels) == 2:
            index = pd.DataFrame(index_data, columns=index_labels).set_index(index_labels).index
        else:
            index = pd.Int64Index(index_data[:, 0], name=index_labels[0])

        column_names = _decode_ascii_list(labels)
        return pd.DataFrame(values, index=index, columns=column_names)

    def _query(self, command, args=None):
        self._send_command(command, args)
        self._check_if_process_still_alive()
        array_num, pickle_data = self._parse_response()

        if isinstance(pickle_data, Exception):
            raise pickle_data

        if array_num == 0:
            return pickle_data

        numpy_arrays = [np.lib.format.read_array(self.proc.stdout) for _ in range(array_num)]

        return pickle_data, numpy_arrays

    def _send_command(self, command, args=None):
        self._check_if_process_still_alive()
        pickle.dump((command, args), self._proc.stdin, protocol=2)
        self._proc.stdin.flush()

    def _parse_response(self):
        pickle_data = b''
        while True:
            line = self._proc.stdout.readline().rstrip() + b'\n'
            pickle_data += line
            if line == b'.\n':
                break
        return pickle.loads(pickle_data, encoding='bytes')

    def __del__(self):
        if self._proc is not None:
            self._send_command('QUIT')

    def _check_if_process_still_alive(self):
        if self._proc.poll() is not None:
            _, error_message = self._proc.communicate()
            self._proc = None

            raise OdbServerError(error_message.decode('ascii'))

def _decode_ascii_list(ascii_list):
    return [item.decode('ascii') for item in ascii_list]



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pytest
import pandas as pd
import numpy as np

import pylife.stress.rainflow


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[-1., 1.], [2, -2]]),
        pd.Series([1., 2.])
    ),
    (
        pd.DataFrame([[-2., 2.], [3, -3]], index=[23, 42]),
        pd.Series([2., 3.], index=[23, 42])
    )
])
def test_rainflow_collective_signal_amplitude_from_to(df, expected):
    df.columns = ['from', 'to']
    expected.name = 'amplitude'
    pd.testing.assert_series_equal(df.rainflow.amplitude, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.rainflow.cycles, expected_cycles)

@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[-1., 1.], [3, -1]]),
        pd.Series([0., 1.])
    ),
    (
        pd.DataFrame([[-2., 2.], [6, -2]], index=[23, 42]),
        pd.Series([0., 2.], index=[23, 42])
    )
])
def test_rainflow_collective_signal_mean_from_to(df, expected):
    df.columns = ['from', 'to']
    expected.name = 'meanstress'
    pd.testing.assert_series_equal(df.rainflow.meanstress, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.rainflow.cycles, expected_cycles)



@pytest.mark.parametrize('df, expected_upper, expected_lower', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[-1., 1.], [3, -1]]),
        pd.Series([1., 3.]), pd.Series([-1., -1])
    ),
    (
        pd.DataFrame([[-2., 2.], [6, -2]], index=[23, 42]),
        pd.Series([2., 6.], index=[23, 42]),
        pd.Series([-2., -2.], index=[23, 42]),
    )
])
def test_rainflow_collective_signal_upper_lower_from_to(df, expected_upper, expected_lower):
    df.columns = ['from', 'to']
    expected_upper.name = 'upper'
    expected_lower.name = 'lower'
    pd.testing.assert_series_equal(df.rainflow.upper, expected_upper)
    pd.testing.assert_series_equal(df.rainflow.lower, expected_lower)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
    ),
    (
        pd.DataFrame([[-1., 1.], [2, -2]]),
        pd.DataFrame([[-2., 2.], [4, -4]]),
    ),
    (
        pd.DataFrame([[-2., 2.], [3, -3]], index=[23, 42]),
        pd.DataFrame([[-4., 4.], [6, -6]], index=[23, 42]),
    )
])
def test_rainflow_collective_from_to_scale_scalar(df, expected):
    df.columns = ['from', 'to']
    expected.columns = ['from', 'to']
    pd.testing.assert_frame_equal(df.rainflow.scale(2.0).to_pandas(), expected)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
    ),
    (
        pd.DataFrame([[-1., 1.], [2, -2]]),
        pd.DataFrame([[1., 3.], [4., 0.]]),
    ),
    (
        pd.DataFrame([[-2., 2.], [3., -3.]], index=[23, 42]),
        pd.DataFrame([[0., 4.], [5., -1.]], index=[23, 42]),
    )
])
def test_rainflow_collective_from_to_shift_scalar(df, expected):
    df.columns = ['from', 'to']
    expected.columns = ['from', 'to']
    pd.testing.assert_frame_equal(df.rainflow.shift(2.0).to_pandas(), expected)


def test_rainflow_collective_from_to_scale_series():
    df = pd.DataFrame([[-2., 2.], [3, -3]],
                      columns=['from', 'to'],
                      index=pd.Index([23, 42], name='foo_index'))

    scale_operand = pd.Series([2, 3, 4], index=pd.Index([6, 7, 8], name='scale_index'))

    expected_index = pd.MultiIndex.from_tuples([
        (23, 6), (23, 7), (23, 8),
        (42, 6), (42, 7), (42, 8),
    ], names=['foo_index', 'scale_index'])
    expected = pd.DataFrame([
        [-4., 4.],
        [-6., 6.],
        [-8., 8.],
        [6., -6.],
        [9., -9.],
        [12., -12.]
    ], columns=['from', 'to'], index=expected_index)

    pd.testing.assert_frame_equal(df.rainflow.scale(scale_operand).to_pandas(), expected)


def test_rainflow_collective_from_to_shift_series():
    df = pd.DataFrame([[-2., 2.], [3, -3]],
                      columns=['from', 'to'],
                      index=pd.Index([23, 42], name='foo_index'))

    shift_operand = pd.Series([2, 3, 4], index=pd.Index([6, 7, 8], name='shift_index'))

    expected_index = pd.MultiIndex.from_tuples([
        (23, 6), (23, 7), (23, 8),
        (42, 6), (42, 7), (42, 8),
    ], names=['foo_index', 'shift_index'])
    expected = pd.DataFrame([
        [0., 4.],
        [1., 5.],
        [2., 6.],
        [5., -1.],
        [6., 0.],
        [7., 1.]
    ], columns=['from', 'to'], index=expected_index)

    pd.testing.assert_frame_equal(df.rainflow.shift(shift_operand).to_pandas(), expected)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([2., 1.])
    ),
    (
        pd.DataFrame([[6., 2.], [4, -3]], index=[23, 42]),
        pd.Series([3., 2.], index=[23, 42])
    )
])
def test_rainflow_collective_signal_amplitude_range_mean(df, expected):
    df.columns = ['range', 'mean']
    expected.name = 'amplitude'
    pd.testing.assert_series_equal(df.rainflow.amplitude, expected)

    expected_cycles = pd.Series(1.0, name='cycles', index=df.index)
    pd.testing.assert_series_equal(df.rainflow.cycles, expected_cycles)


@pytest.mark.parametrize('df, expected_upper, expected_lower', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([3., 1.]), pd.Series([-1., -1])
    ),
    (
        pd.DataFrame([[6., 2.], [4, -3]], index=[23, 42]),
        pd.Series([5., -1.], index=[23, 42]),
        pd.Series([-1., -5.], index=[23, 42]),
    )
])
def test_rainflow_collective_signal_upper_lower_range_mean(df, expected_upper, expected_lower):
    df.columns = ['range', 'mean']
    expected_upper.name = 'upper'
    expected_lower.name = 'lower'
    pd.testing.assert_series_equal(df.rainflow.upper, expected_upper)
    pd.testing.assert_series_equal(df.rainflow.lower, expected_lower)


@pytest.mark.parametrize('df, expected', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([1., 0.])
    ),
    (
        pd.DataFrame([[6., 2.], [4, -3]], index=[23, 42]),
        pd.Series([2., -3.], index=[23, 42])
    )
])
def test_rainflow_collective_signal_mean_range_mean(df, expected):
    df.columns = ['range', 'mean']
    expected.name = 'meanstress'
    pd.testing.assert_series_equal(df.rainflow.meanstress, expected)


@pytest.mark.parametrize('df, expected_amplitude, expected_mean', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([4., 2.]), pd.Series([2., 0.])
    ),
    (
        pd.DataFrame([[6., 2.], [4., -3.]], index=[23, 42]),
        pd.Series([6., 4.], index=[23, 42]),
        pd.Series([4., -6.], index=[23, 42]),
    )
])
def test_rainflow_collective_signal_mean_range_scale_scalar(df, expected_amplitude, expected_mean):
    df.columns = ['range', 'mean']
    expected_amplitude.name = 'amplitude'
    expected_mean.name = 'meanstress'
    scaled = df.rainflow.scale(2.0)
    pd.testing.assert_series_equal(scaled.amplitude, expected_amplitude)
    pd.testing.assert_series_equal(scaled.meanstress, expected_mean)


@pytest.mark.parametrize('df, expected_amplitude, expected_mean', [
    (
        pd.DataFrame(columns=[1, 2], dtype=np.float64),
        pd.Series(dtype=np.float64), pd.Series(dtype=np.float64)
    ),
    (
        pd.DataFrame([[4., 1.], [2., 0.]]),
        pd.Series([2., 1.]), pd.Series([3., 2.])
    ),
    (
        pd.DataFrame([[6., 2.], [4., -3.]], index=[23, 42]),
        pd.Series([3., 2.], index=[23, 42]),
        pd.Series([4., -1.], index=[23, 42]),
    )
])
def test_rainflow_collective_signal_mean_range_shift_scalar(df, expected_amplitude, expected_mean):
    df.columns = ['range', 'mean']
    expected_amplitude.name = 'amplitude'
    expected_mean.name = 'meanstress'
    scaled = df.rainflow.shift(2.0)
    pd.testing.assert_series_equal(scaled.amplitude, expected_amplitude)
    pd.testing.assert_series_equal(scaled.meanstress, expected_mean)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pandas as pd
import pytest
from pylife.strength import sn_curve


def load_index():
    nCode_Xbinsize = 62.375
    nCode_XMax = 468.8125
    nCode_XMin = 32.1875

    return pd.interval_range(start=nCode_XMin-nCode_Xbinsize/2,
                             end=nCode_XMax+nCode_Xbinsize/2, periods=8, name="range")


def nCode_reference_results():
    elementar = pd.DataFrame([[2.057849E-06, 1.868333E-06, 3.067487E-06],
                              [3.039750E-05, 3.894329E-05, 1.999014E-08],
                              [1.507985E-05, 4.829155E-05, 0.000000E+00],
                              [6.434856E-06, 1.543643E-05, 0.000000E+00],
                              [6.296737E-06, 1.357734E-05, 0.000000E+00],
                              [1.751881E-06, 1.379607E-05, 0.000000E+00],
                              [0.000000E+00, 1.023415E-05, 0.000000E+00],
                              [0.000000E+00, 7.548524E-07, 0.000000E+00]],
                             index=load_index()).stack()
    miner_haibach = pd.DataFrame([[2.293973E-08, 2.083222E-08, 3.419927E-08],
                                  [7.414563E-05, 9.499395E-05, 4.876799E-08],
                                  [4.631857E-04, 1.483300E-03, 0.000000E+00],
                                  [4.779392E-04, 1.146517E-03, 0.000000E+00],
                                  [6.006944E-04, 1.295247E-03, 0.000000E+00],
                                  [2.041326E-04, 1.607545E-03, 0.000000E+00],
                                  [0.000000E+00, 1.408691E-03, 0.000000E+00],
                                  [0.000000E+00, 1.198482E-04, 0.000000E+00]],
                                 index=load_index()).stack()
    original = pd.DataFrame([[0.00000000E+00, 0.00000000E+00, 0.00000000E+00],
                             [0.00000000E+00, 0.00000000E+00, 0.00000000E+00],
                             [2.08681840E-05, 6.68280320E-05, 0.00000000E+00],
                             [1.73881920E-05, 4.17121220E-05, 0.00000000E+00],
                             [2.80698130E-05, 6.05255350E-05, 0.00000000E+00],
                             [1.16511320E-05, 9.17526660E-05, 0.00000000E+00],
                             [0.00000000E+00, 9.49790820E-05, 0.00000000E+00],
                             [0.00000000E+00, 9.32071420E-06, 0.00000000E+00]],
                            index=load_index()).stack()
    elementar.name = 'damage'
    miner_haibach.name = 'damage'
    original.name = 'damage'
    return [elementar, miner_haibach, original]


material = pd.DataFrame(index=['k_1', 'ND_50', 'SD_50'],
                        columns=['elementar', 'MinerHaibach', 'original'],
                        data=[[4, 5, 6], [4e7, 1e6, 1e8], [100, 90, 75]])

loads = pd.DataFrame([[1.227E5, 1.114E5, 1.829E5], [2.433E4, 3.117E4, 16],
                      [1591, 5095, 0], [178, 427, 0], [64, 138, 0], [8, 63, 0],
                      [0, 24, 0], [0, 1, 0]],
                     index=load_index()).stack()


@pytest.mark.parametrize('method, expected', zip(material, nCode_reference_results()))
def test_calc_damage(method, expected):
    with pytest.warns(DeprecationWarning):
        damage_calc = sn_curve.FiniteLifeCurve(**material[method])
    damage = damage_calc.calc_damage(loads, method=method)
    pd.testing.assert_series_equal(damage, expected, rtol=0.001, atol=0)


def test_calc_damage_index_name():
    foo_loads = loads.copy()
    foo_loads.index.name = 'foo'
    with pytest.warns(DeprecationWarning):
        damage = sn_curve.FiniteLifeCurve(**material['elementar']).calc_damage(foo_loads,
                                                                               method='elementar',
                                                                               index_name='foo')

    assert damage.index.names[0] == 'foo'



# -*- coding: utf-8 -*-
"""
Created on Tue Dec  8 11:26:26 2020

@author: KRD2RNG
"""

import pyvista as pv
from pyvista import examples
import numpy as np

import pylife.vmap
import pylife.stress.equistress
#%%
filename = "../plate_with_hole.vmap"

vm_mesh = pylife.vmap.VMAPImport(filename)
pyLife_mesh = df = (vm_mesh.mesh_coords('1')
                    .join(vm_mesh.variable('1', 'STATE-2', 'STRESS_CAUCHY'))
                    .join(vm_mesh.variable('1', 'STATE-2', 'DISPLACEMENT')))
pyLife_mesh['mises'] = pyLife_mesh.equistress.mises()
pyLife_nodes = pyLife_mesh.groupby('node_id').mean()
mesh = pv.PolyData(pyLife_nodes[['x', 'y', 'z']].values)
mesh.plot(point_size=10, screenshot=False)
mesh.point_arrays["mises"] = pyLife_nodes["mises"].values
mesh.plot(scalars="mises")



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np


def find_turns(samples):
    """Find the turning points in a sample chunk.

    Parameters
    ----------
    samples : 1D numpy.ndarray
        the sample chunk

    Returns
    -------
    index : 1D numpy.ndarray
        the indeces where sample has a turning point
    turns : 1D numpy.ndarray
        the values of the turning points

    Notes
    -----
    In case of plateaus i.e. multiple directly neighbored samples with exactly
    the same values, building a turning point together, the first sample of the
    plateau is indexed.

    """
    def plateau_turns(diffs):
        plateau_turns = np.zeros_like(diffs, dtype=np.bool_)[1:]
        duplicates = np.array(diffs == 0, dtype=np.int8)

        if duplicates.any():
            edges = np.diff(duplicates)
            dups_starts = np.where(edges > 0)[0]
            dups_ends = np.where(edges < 0)[0]
            if len(dups_starts) and len(dups_ends):
                cut_ends = dups_ends[0] < dups_starts[0]
                cut_starts = dups_starts[-1] > dups_ends[-1]
                if cut_ends:
                    dups_ends = dups_ends[1:]
                if cut_starts:
                    dups_starts = dups_starts[:-1]
                plateau_turns[dups_starts[np.where(diffs[dups_starts] * diffs[dups_ends+1] < 0)]] = True

        return plateau_turns

    diffs = np.diff(samples)
    peak_turns = diffs[:-1] * diffs[1:] < 0.0

    index = np.where(np.logical_or(peak_turns, plateau_turns(diffs)))[0] + 1

    return index, samples[index]


class AbstractDetector:
    """The common base class for rainflow detectors.

    Subclasses implementing a specific rainflow counting algorithm are supposed
    to implement a method ``process()`` that takes the signal samples as a
    parameter, and reports all the hysteresis loop limits to ``self._recorder``
    using its ``record_values()`` method of.

    Some detectors also report the index of the loop limiting samples to the
    recorder using its ``record_index()`` method. Those detectors should also
    report the size of each processed sample chunk to the recorder using
    ``report_chunk()``.

    The ``process()`` method is supposed return ``self`` and to be implemented
    in a way, that the result is independent of the sample chunksize, so
    ``dtor.process(signal)`` should be equivalent to
    ``dtor.process(signal[:n]).process(signal[n:])`` for any 0 < n < signal
    length.

    Should usually only be instantiated by a sublacsse's ``__init__()`` using
    ``super().__init__()``.
    """

    def __init__(self, recorder):
        """Instantiate an AbstractDetector.

        Parameters
        ----------
        recorder : subclass of :class:`.AbstractRecorder`
            The recorder that the detector will report to.
        """
        self._sample_tail = np.array([])
        self._recorder = recorder
        self._head_index = 0
        self._residual_index = np.array([0], dtype=np.int64)
        self._residuals = np.array([])

    @property
    def residuals(self):
        """The residual turning points of the time signal so far.

        The residuals are the loops not (yet) closed.
        """
        return self._residuals

    @property
    def residual_index(self):
        """The index of the residual turning points of the time signal so far.
        """
        return np.append(self._residual_index, self._head_index - 1)

    @property
    def recorder(self):
        """The recorder instance the detector is reporting to.
        """
        return self._recorder

    def _new_turns(self, samples):
        """Provide new turning points for the next chunk

        Parameters
        ----------
        samples : 1-D array of float
            The samples of the chunk to be processed

        Returns
        -------
        turn_index : 1-D array of int
            The global index of the turning points of the chunk to be processed
        turn_values : 1-D array of float
            The values of the turning points

        Notes
        -----

        This method can be called by the ``process()`` implementation of
        subclasses. The sample tail i.e. the samples after the last turning
        point of the chunk are stored and prepended to the samples of the next
        call.
        """
        sample_len = len(samples)
        samples = np.concatenate((self._sample_tail, samples))
        turn_index, turn_values = find_turns(samples)
        if turn_index.size > 0:
            old_sample_tail_length = len(self._sample_tail)
            self._sample_tail = samples[turn_index[-1]:]
            turn_index += self._head_index - old_sample_tail_length
        else:
            self._sample_tail = samples

        self._head_index += sample_len

        return turn_index, turn_values


class AbstractRecorder:
    """A common base class for rainflow recorders.

    Subclasses implementing a rainflow recorder are supposed to implement the
    following methods:

    * ``record_values()``
    * ``record_index()``
    """

    def __init__(self):
        """Instantiate an AbstractRecorder."""
        self._chunks = np.array([], dtype=np.int64)

    @property
    def chunks(self):
        """The limits index of the chunks processed so far.

        Note
        ----
        The first chunk limit is the length of the first chunk, so identical to
        the index to the first sample of the second chunk, if a second chunk
        exists.
        """
        return self._chunks

    def report_chunk(self, chunk_size):
        """Report a chunk.

        Parameters
        ----------
        chunk_size : int
            The length of the chunk previously processed by the detector.

        Note
        ----
        Should be called by the detector after the end of ``process()``.
        """
        self._chunks = np.append(self._chunks, chunk_size)

    def chunk_local_index(self, global_index):
        """Transform the global index to an index valid in a certain chunk.

        Parameters
        ----------
        global_index : array-like int
            The global index to be transformed.

        Returns
        -------
        chunk_number : array of ints
            The number of the chunk the indexed sample is in.
        chunk_local_index : array of ints
            The index of the sample in its chunk.
        """
        chunk_index = np.insert(np.cumsum(self._chunks), 0, 0)
        chunk_num = np.searchsorted(chunk_index, global_index, side='right') - 1

        return chunk_num, global_index - chunk_index[chunk_num]

    def record_values(self, value_from, value_to):
        """Record hysteresis loop values to the recorder.

        Parameters
        ----------
        value_from : float
            The sample value where the hysteresis loop starts from.

        value_to : float
            The sample value where the hysteresis loop goes to and turns back from.

        Note
        ----
        Default implementation does nothing. To be implemented by recorders
        interested in the hysteresis loop values.
        """
        pass

    def record_index(self, index_from, index_to):
        """Record hysteresis loop index to the recorder.

        Parameters
        ----------
        index_from : float
            The sample index where the hysteresis loop starts from.

        index_to : float
            The sample index where the hysteresis loop goes to and turns back from.

        Note
        ----
        Default implementation does nothing. To be implemented by recorders
        interested in the hysteresis loop values.
        """
        pass
